function add(a, b) {
  return b + a;
}

/**
 * Curries the given function `f`.
 *
 * Currying turns a function that takes multiple arguments into a function that
 * can be applied to one or more values to give a partially applied function.
 *
 * @param {Function} f The function to curry.
 * @returns {Function} A curried function.
 * @example
 *
 * import { curry } from 'fkit'
 * const add = curry((a, b) => a + b)
 * const add2 = add(2)
 * add2(1) // 3
 */
function curry(f) {
  var arity = f.length;

  var g = function g(oldArgs) {
    return function () {
      for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      var newArgs = oldArgs.concat(args.length > 0 ? args : undefined); // If we have enough args, then apply the function.

      return newArgs.length >= arity ? f.apply(undefined, newArgs) : g(newArgs);
    };
  };

  return arity <= 1 ? f : g([]);
}

/**
 * Returns the sum of two numbers.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Number} The result of `b + a`.
 * @example
 *
 * import { add } from 'fkit'
 * add(1, 2) // 3
 */

var add$1 = curry(add);

/**
 * Returns a function that always returns the same value, regardless of the
 * arguments.
 *
 * @param c The constant value.
 * @returns {Function} A function that always returns the value `c`.
 * @example
 *
 * import { always } from 'fkit'
 * always(1)() // 1
 */
function always(c) {
  return function () {
    return c;
  };
}

function branch(p, f, g, a) {
  return p(a) ? f(a) : g(a);
}

/**
 * Branches execution based on a predicate function.
 *
 * If `p(a)` is true then `f` is applied to `a`, otherwise `g` is applied to
 * `a`.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Function} f The function to apply when the predicate is satisfied.
 * @param {Function} g The function to apply when the predicate is unsatisfied.
 * @param a The value.
 * @returns The result.
 * @example
 *
 * import { branch, gt } from 'fkit'
 * const big = a => a + ' is a big number'
 * const small = a => a + ' is a small number'
 * const f = branch(gt(10), big, small)
 * f(10) // small number
 * f(11) // big number
 */

var branch$1 = curry(branch);

/**
 * The identity function.
 *
 * @param a The value to return.
 * @returns The value `a` unchanged.
 * @example
 *
 * import { id } from 'fkit'
 * id(1) // 1
 */
function id(a) {
  return a;
}

/**
 * Returns true if `a` is a string.
 *
 * @private
 */
function isString(a) {
  return typeof a === 'string';
}

/**
 * Flattens the list of `as`.
 *
 * @private
 */
function flatten(as) {
  return as.reduce(function (a, b) {
    return a.concat(b);
  }, []);
}

var slice = Array.prototype.slice;
/**
 * Converts a function to a variadic function.
 *
 * The last named parameter will be given an array of arguments.
 *
 * @param {Function} f The function.
 * @returns {Function} A function that wraps the function `f` to accept any
 * number of arguments.
 * @example
 *
 * import { variadic } from 'fkit'
 * function f (head, tail) { ... }
 * variadic(f)(1, 2, 3) // f(1, [2, 3])
 */

function variadic(f) {
  var _this = this;

  var arity = f.length;

  if (arity < 1) {
    return f;
  } else if (arity === 1) {
    return function () {
      for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      var newArgs = args.length === 1 ? flatten(args) : args;
      return f.call(_this, newArgs);
    };
  } else {
    return function () {
      for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        args[_key2] = arguments[_key2];
      }

      var numMissingArgs = Math.max(arity - args.length - 1, 0);
      var missingArgs = new Array(numMissingArgs);
      var namedArgs = slice.call(args, 0, arity - 1);
      var variadicArgs = slice.call(args, f.length - 1);
      return f.apply(_this, namedArgs.concat(missingArgs).concat([variadicArgs]));
    };
  }
}

/**
 * Composes a list of functions.
 *
 * @function
 * @param {Array} fs The functions to compose.
 * @returns {Function} A function that is the composition of the list of
 * functions `fs`.
 * @example
 *
 * import { compose } from 'fkit'
 * compose(f, g, h)(a) // f(g(h(a)))
 */

var compose = variadic(function (fs) {
  return function (a) {
    return fs.reduceRight(function (a, f) {
      return f(a);
    }, a);
  };
});

function append(a, bs) {
  return isString(bs) ? bs + a : bs.concat([a]);
}

/**
 * Appends a value to a list.
 *
 * @function
 * @param a The value to append.
 * @param {Array|String} bs The list.
 * @returns {Array|String} A list that contains the value `a` appended to the
 * list of `bs`.
 * @example
 *
 * import { append } from 'fkit'
 * append(3, [1, 2]) // [1, 2, 3]
 * append('o', 'fo') // 'foo'
 */

var append$1 = curry(append);

/**
 * Swaps the order of the arguments to a function.
 *
 * @param {Function} f The function.
 * @returns {Function} A new function which calls `f` with the arguments swapped.
 * @example
 *
 * import { flip } from 'fkit'
 * function f (a, b) { ... }
 * const g = flip(f)
 * g(1, 2) // f(2, 1)
 */
function flip(f) {
  return function (a, b) {
    return f(b, a);
  };
}

/**
 * Returns a list that contains the elements in the list of `as` concatenated
 * with the starting value `s`.
 *
 * @private
 */

function concatWith(s, as) {
  return Array.from(flatten(as)).reduce(flip(append$1), s);
}

/**
 * Returns true if `as` is an array.
 *
 * @private
 */
function isArray(as) {
  return Array.isArray(as);
}

/**
 * Returns true if the list of `as` is an array of strings, false otherwise.
 *
 * @private
 */

function isArrayOfStrings(as) {
  return isArray(as) && as.length > 0 && as.reduce(function (a, b) {
    return a && isString(b);
  }, true);
}

/**
 * Returns an empty monoid of `as`.
 *
 * @private
 */

function mempty(as) {
  return isString(as) || isArrayOfStrings(as) ? '' : [];
}

/**
 * Concatenates multiple lists into a single list.
 *
 * @function
 * @param {Array|String} as The lists to concatenate.
 * @returns {Array|String} A list that contains the concatenated elements in
 * the list of `as`.
 * @example
 *
 * import { concat } from 'fkit'
 * concat([1], [2, 3], [4, 5, 6]) // [1, 2, 3, 4, 5, 6]
 * concat('f', 'oo', 'bar') // 'foobar'
 */

var concat = variadic(function (as) {
  return concatWith(mempty(as), as);
});

/**
 * Flattens any strings in the list of `as`.
 *
 * @private
 */

function flattenStrings(as) {
  if (isArrayOfStrings(as)) {
    return concat(as);
  } else if (isArray(as)) {
    return as.map(flattenStrings);
  } else {
    return as;
  }
}

function concatMap(f, as) {
  var bs = Array.from(as).map(compose(flattenStrings, f));
  var cs = bs.length > 0 ? bs : as;
  return concatWith(mempty(cs), bs);
}

function filter(p, as) {
  var f = branch$1(p, id, always(''));
  return isString(as) ? concatMap(f, as) : as.filter(p);
}

function all(p, as) {
  return filter(p, as).length === as.length;
}

/**
 * Determines whether all elements in a list satisfy a predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns {Boolean} `true` if all elements of the list of `as` satisfy the
 * predicate, `false` otherwise.
 * @example
 *
 * import { all, eq, gt } from 'fkit'
 *
 * all(gt(1), [1, 2, 3]) // false
 * all(gt(1), [2, 3]) // true
 * all(gt(1), [3]) // true
 *
 * all(eq('o'), 'foo') // false
 * all(eq('o'), 'oo') // true
 * all(eq('o'), 'o') // true
 */

var all$1 = curry(all);

function and(a, b) {
  return Boolean(b && a);
}

/**
 * Returns the logical AND of two values.
 *
 * @function
 * @param a The first value.
 * @param b The second value.
 * @returns {Boolean} The result of `b && a`.
 * @example
 *
 * import { and } from 'fkit'
 * and(true, true) // true
 * and(0, 1) // false
 * and('', 'foo') // false
 */

var and$1 = curry(and);

function any(p, as) {
  return filter(p, as).length > 0;
}

/**
 * Determines whether any elements in a list satisfy a predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns {Boolean} `true` if any elements in the list of `as` satisfy the
 * predicate function `p`, `false` otherwise.
 * @example
 *
 * import { any, eq, gt } from 'fkit'
 *
 * any(gt(1), [1, 2, 3]) // true
 * any(gt(1), [1, 2]) // true
 * any(gt(1), [1]) // false
 *
 * any(eq('o'), 'foo') // true
 * any(eq('o'), 'fo') // true
 * any(eq('o'), 'f') // false
 */

var any$1 = curry(any);

function apply(f, a) {
  return f(a);
}

/**
 * Applies a function to a value.
 *
 * @function
 * @param {Function} f The function to apply.
 * @param a The value.
 * @returns The result of `f(a)`.
 * @example
 *
 * import { apply } from 'fkit'
 * const sayHi = a => ['Hi', a, '!'].join(' ')
 * apply(sayHi, 'jane') // Hi Jane!
 */

var apply$1 = curry(apply);

function apply2(f, a, b) {
  return f(a, b);
}

/**
 * Applies a function to two values.
 *
 * @function
 * @param {Function} f The function to apply.
 * @param a The first value.
 * @param b The second value.
 * @returns The result of `f(a, b)`.
 * @example
 *
 * import { apply2 } from 'fkit'
 * const sayHi = (a, b) => ['Hi', a, b, '!'].join(' ')
 * apply2(sayHi, 'Jane', 'Appleseed') // Hi Jane Appleseed!
 */

var apply2$1 = curry(apply2);

function apply3(f, a, b, c) {
  return f(a, b, c);
}

/**
 * Applies a function to three values.
 *
 * @function
 * @param {Function} f The function to apply.
 * @param a The first value.
 * @param b The second value.
 * @param c The third value.
 * @returns The result of `f(a, b, c)`.
 * @example
 *
 * import { apply3 } from 'fkit'
 * const sayHi = (a, b, c) => ['Hi', a, b, c, '!'].join(' ')
 * apply3(sayHi, 'Ms', 'Jane', 'Appleseed') // Hi Ms Jane Appleseed!
 */

var apply3$1 = curry(apply3);

function applyMethod(k, a, o) {
  return o[k](a);
}

/**
 * Applies a method to a value.
 *
 * @function
 * @param {String} k The method to apply.
 * @param a The value.
 * @param {Object} o The object.
 * @returns The result of the method `k` of object `o` applied to the value
 * `a`.
 * @example
 *
 * import { applyMethod } from 'fkit'
 * const person = { sayHi: a => ['Hi', a, '!'].join(' ') }
 * applyMethod('sayHi', 'Jane', person) // Hi Jane!
 */

var applyMethod$1 = curry(applyMethod);

function applyMethod2(k, a, b, o) {
  return o[k](a, b);
}

/**
 * Applies a method to two values.
 *
 * @function
 * @param {String} k The method to apply.
 * @param a The first value.
 * @param b The second value.
 * @param {Object} o The object.
 * @returns The result of the method `k` of object `o` applied to the values
 * `a` and `b`.
 * @example
 *
 * import { applyMethod2 } from 'fkit'
 * const person = { sayHi: (a, b) => ['Hi', a, b, '!'].join(' ') }
 * applyMethod2('sayHi', 'Jane', 'Appleseed', person) // Hi Jane Appleseed!
 */

var applyMethod2$1 = curry(applyMethod2);

function applyMethod3(k, a, b, c, o) {
  return o[k](a, b, c);
}

/**
 * Applies a method to three values.
 *
 * @function
 * @param {String} k The method to apply.
 * @param a The first value.
 * @param b The second value.
 * @param c The third value.
 * @param {Object} o The object.
 * @returns The result of the method `k` of object `o` applied to the values
 * `a`, `b`, and `c`.
 * @example
 *
 * import { applyMethod3 } from 'fkit'
 * const person = { sayHi: (a, b, c) => ['Hi', a, b, c, '!'].join(' ') }
 * applyMethod3('sayHi', 'Ms', 'Jane', 'Appleseed', person) // Hi Ms Jane Appleseed!
 */

var applyMethod3$1 = curry(applyMethod3);

function applyRight(a, f) {
  return f(a);
}

/**
 * Applies a function to a value. This is similar to `apply`, however the order
 * of the arguments is flipped.
 *
 * @function
 * @param a The value.
 * @param {Function} f The function to apply.
 * @returns The result of `f(a)`.
 * @example
 *
 * import { applyRight } from 'fkit'
 * function sayHi (a) { return ['Hi', a, '!'].join(' ') }
 * applyRight('Jane', sayHi) // Hi Jane!
 */

var applyRight$1 = curry(applyRight);

/**
 * Creates a new array.
 *
 * @param {Number} n The length of the array.
 * @returns {Array} An array of length `n`.
 * @example
 *
 * import { array } from 'fkit'
 * array(3) // [undefined, undefined, undefined]
 */
function array(n) {
  return Array.apply(null, Array(n));
}

function between(a, b, n) {
  return a <= n && b >= n;
}

/**
 * Determines whether a number is between two numbers.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @param {Number} n The number to test.
 * @returns {Number} `true` if the number `n` is between the numbers `a` and
 * `b` (i.e. a <= n <= b), `false` otherwise.
 * @example
 *
 * import { between } from 'fkit'
 * between(1, 3, 2) // true
 * between(1, 3, 4) // false
 */

var between$1 = curry(between);

/**
 * Converts a function of any arity to a binary function.
 *
 * @param {Function} f The function.
 * @returns {Function} A function that wraps the function `f` to accept only
 * two arguments.
 * @example
 *
 * import { binary } from 'fkit'
 * function f (a, b) { ... }
 * const g = binary(f)
 * g(1, 2, 3) // f(1, 2)
 */

function binary(f) {
  return f.length === 2 ? f : apply2$1(f);
}

/**
 * Determines if a list is empty.
 *
 * @param {Array|String} as The list.
 * @returns {Boolean} `true` if the list of `as` is empty, `false` otherwise.
 * @example
 *
 * import { empty } from 'fkit'
 *
 * empty([]) // true
 * empty([1, 2, 3]) // false
 *
 * empty('') // true
 * empty('foo') // false
 */
function empty(as) {
  return as.length === 0;
}

/**
 * Gets the first element in a list.
 *
 * @param {Array|String} as The list.
 * @returns The first element in the list of `as`, or `undefined` if the list
 * is empty.
 * @example
 *
 * import { head } from 'fkit'
 * head([1, 2, 3]) // 1
 * head('foo') // 'f'
 */
function head(as) {
  return as[0];
}

function pair(a, b) {
  return [a, b];
}

/**
 * Creates a new ordered pair.
 *
 * @function
 * @param a The first value.
 * @param b The second value.
 * @returns {Array} An ordered pair with the values `a` and `b`.
 * @example
 *
 * import { pair } from 'fkit'
 * pair(1, 2) // [1, 2]
 * pair('a', 'b') // ['a', 'b']
 */

var pair$1 = curry(pair);

/**
 * Get the elements after the first element in a list.
 *
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the elements after the first
 * element in the list of `as`.
 * @example
 *
 * import { tail } from 'fkit'
 * tail([1, 2, 3]) // [2, 3]
 * tail('foo') // 'oo'
 */
function tail(as) {
  return as.slice(1);
}

function map(f, as) {
  return Array.from(as).map(f);
}

function cartesian(as, bs) {
  return empty(as) ? [] : concat(map(pair$1(head(as)), bs), cartesian(tail(as), bs));
}

/**
 * Calculates the [Cartesian
 * product](https://en.wikipedia.org/wiki/Cartesian_product) of two lists.
 *
 * @function
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array} Returns a list that contains all the ordered pairs `[a, b]`
 * in the lists of `as` and `bs`.
 * @example
 *
 * import { cartesian } from 'fkit'
 * cartesian([1, 2], [3, 4]) // [[1, 3], [1, 4], [2, 3], [2, 4]]
 * cartesian('ab', 'cd') // [['a', 'c'], ['a', 'd'], ['b', 'c'], ['b', 'd']]
 */

var cartesian$1 = curry(cartesian);

function eq(a, b) {
  return b === a;
}

/**
 * Determines whether two values are equal.
 *
 * @function
 * @param a The first value.
 * @param b The second value.
 * @returns {Boolean} `true` if the value `a` is strictly not equal (`===`) to
 * the value `b`, false otherwise.
 * @example
 *
 * import { eq } from 'fkit'
 * eq(1, 1) // true
 * eq(1, 2) // false
 */

var eq$1 = curry(eq);

/**
 * Gets the last element in a list.
 *
 * @param {Array|String} as The list.
 * @returns The last element in the list of `as`, or `undefined` if the list is empty.
 * @example
 *
 * import { last } from 'fkit'
 * last([1, 2, 3]) // 3
 * last('foo') // 'o'
 */
function last(as) {
  return as[as.length - 1];
}

function dropWhile(p, as) {
  var s = isString(as) ? '' : [];
  var m = as.length;
  var n = 0;

  while (n < m && p(as[n])) {
    n++;
  }

  for (var i = n; i < m; i++) {
    s = s.concat(as[i]);
  }

  return s;
}

function takeWhile(p, as) {
  var s = isString(as) ? '' : [];
  var n = as.length;

  for (var i = 0; i < n && p(as[i]); i++) {
    s = s.concat(as[i]);
  }

  return s;
}

function span(p, as) {
  return [takeWhile(p, as), dropWhile(p, as)];
}

/**
 * Splits a list using a predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns {Array} A list that contains the elements in the list of `as` split
 * into a pair of lists: a prefix of elements that satisfy the predicate
 * function `p` and the remainder of the list.
 * @example
 *
 * import { lt, neq, span } from 'fkit'
 * span(lt(3), [1, 2, 3]) // [[1, 2], [3]]
 * span(neq(o), 'foo') // ['f', 'oo']
 */

var span$1 = curry(span);

function prepend(a, bs) {
  return isString(bs) ? a + bs : [a].concat(bs);
}

function chunkBy(c, as) {
  var b = head(as);
  var bs = span$1(function (a) {
    return c(a, b);
  }, tail(as));
  return empty(as) ? [] : prepend(prepend(b, head(bs)), chunkBy(c, last(bs)));
}

/**
 * Chunks contiguous values in a list using a comparator function.
 *
 * The comparator function `f` compares two values, `a` and `b`. If the values
 * are both considered to be in the same chunk, then the comparator function
 * should return `true`. Otherwise it should return `false`.
 *
 * @function
 * @param {Function} c The comparator function.
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the values in the list of `as`
 * chunked into sublists that satisfy the comparator function `c`.
 * @example
 *
 * import { chunkBy } from 'fkit'
 * chunkBy((a, b) => a === b, [1, 2, 2, 3, 3, 3]) // [[1], [2, 2], [3, 3, 3]]
 * chunkBy((a, b) => a === b, 'Mississippi') // ['M', 'i', 'ss', 'i', 'ss', 'i', 'pp', 'i']
 */

var chunkBy$1 = curry(chunkBy);

/**
 * Chunks contiguous values in a list.
 *
 * This is a special case of the `chunkBy` function where the values are compared
 * using the strict equality `===` operator.
 *
 * @function
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the values in the list of `as`
 * chunked into sublists of equal values.
 * @example
 *
 * import { chunk } from 'fkit'
 * chunk([1, 2, 2, 3, 3, 3]) // [[1], [2, 2], [3, 3, 3]]
 * chunk('Mississippi') // ['M', 'i', 'ss', 'i', 'ss', 'i', 'pp', 'i']
 */

var chunk = chunkBy$1(eq$1);

function max(a, b) {
  return b > a ? b : a;
}

function min(a, b) {
  return a > b ? b : a;
}

function clamp(a, b, n) {
  return max(a, min(n, b));
}

/**
 * Clamps a number between two values.
 *
 * @function
 * @param {Number} a The minimum number.
 * @param {Number} b The maximum number.
 * @param {Number} n The number to clamp.
 * @returns {Number} The result of clamping the number `n` between the numbers
 * `a` and `b` (i.e. `a <= n <= b`).
 * @example
 *
 * import { clamp } from 'fkit'
 * clamp(1, 3, 0) // 1
 * clamp(1, 3, 2) // 2
 * clamp(1, 3, 4) // 3
 */

var clamp$1 = curry(clamp);

function compare(a, b) {
  if (a > b) {
    return 1;
  } else if (a < b) {
    return -1;
  } else {
    return 0;
  }
}

/**
 * Compares two values using natural ordering.
 *
 * This function compares two elements, `a` and `b`. If `a` is greater than
 * `b`, then it returns `1`. If `a` is less than `b`, then it returns `-1`. If
 * both elements are equal, then it returns `0`.
 *
 * @function
 * @param a The first value.
 * @param b The second value.
 * @returns {Number} The result.
 * @example
 *
 * import { compare } from 'fkit'
 * compare(2, 1) // 1
 * compare(1, 2) // -1
 * compare(2, 2) // 0
 */

var compare$1 = curry(compare);

/**
 * Maps a function, that returns a list, over a list and concatenates the
 * results.
 *
 * @function
 * @param {Function} f The function to apply to each value in the list. It must
 * also return a list.
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as`, mapped with the function `f`, and concatenated together.
 * @example
 *
 * import { concatMap } from 'fkit'
 * concatMap(a => [a, 0], [1, 2, 3]) // [1, 0, 2, 0, 3, 0]
 * concatMap(a => [a, '-'], 'foo') // 'f-o-o-'
 */

var concatMap$1 = curry(concatMap);

function extend(target, objects) {
  objects.forEach(function (object) {
    Object.getOwnPropertyNames(object).forEach(function (property) {
      target[property] = object[property];
    });
  });
  return target;
}
/**
 * Creates a copy of one or more objects.
 *
 * Properties with the same key will take precedence from right to left. The
 * copy will have the *same* prototype as the *first* object in the list.
 *
 * @function
 * @param {Array} os The objects to copy.
 * @returns {Object} A copy of the objects in the list of `os`.
 * @example
 *
 * import { copy } from 'fkit'
 * const person = { name: 'Jane', age: 20, city: 'Melbourne' }
 * copy(person, { name: 'Steve' }) // { name: 'Steve', age: 20, city: 'Melbourne' }
 */


var copy = variadic(function (o, ps) {
  return extend(Object.create(Object.getPrototypeOf(o)), [o].concat(ps));
});

/**
 * Decrements a number.
 *
 * @param {Number} a The number.
 * @returns {Number} The result of `a - 1`.
 * @example
 *
 * import { dec } from 'fkit'
 * dec(1) // 0
 */
function dec(a) {
  return a - 1;
}

function nubBy(f, as) {
  var a = head(as);
  return empty(as) ? mempty(as) : prepend(a, nubBy(f, filter(function (b) {
    return !f(a, b);
  }, tail(as))));
}

/**
 * Removes duplicate elements from a list using a comparator function.
 *
 * The comparator function compares two elements, `a` and `b`. If the elements
 * are both considered to equal, then the comparator function should return
 * `true`. Otherwise it should return `false`.
 *
 * @function
 * @param {Function} f The comparator function.
 * @param {Array|String} as The list.
 * @returns {Array|String} A list with all duplicate elements removed from the
 * list of `bs`.
 * @example
 *
 * import { nubBy } from 'fkit'
 * nubBy((a, b) => a === b, [1, 2, 2, 3, 3, 3]) // [1, 2, 3]
 * nubBy((a, b) => a === b, 'abbccc') // 'abc'
 */

var nubBy$1 = curry(nubBy);

/**
 * Removes duplicate elements from a list.
 *
 * This is a special case of the `nubBy` function where the elements are compared
 * using the strict equality `===` operator.
 *
 * @function
 * @param {Array|String} as The list.
 * @returns {Array|String} A list with all duplicate elements removed from the list of `as`.
 * @example
 *
 * import { nub } from 'fkit'
 * nub([1, 2, 2, 3, 3, 3]) // [1, 2, 3]
 * nub('abbccc') // 'abc'
 */

var nub = nubBy$1(eq$1);

function removeBy(f, a, bs_) {
  var b = head(bs_);
  var bs = tail(bs_);
  return empty(bs_) ? mempty(bs_) : f(a, b) ? bs : prepend(b, removeBy(f, a, bs));
}

/**
 * Removes the first occurance of an element from a list using a comparator
 * function.
 *
 * The comparator function `f` compares two elements, `a` and `b`. If the
 * elements are both considered to equal, then the comparator function should
 * return `true`. Otherwise it should return `false`.
 *
 * @function
 * @param {Function} f The comparator function.
 * @param a The value to find.
 * @param {Array|String} bs The list.
 * @returns {Array|String} A list with the first occurance of the element `a`
 * removed from the list of `bs`.
 * @example
 *
 * import { removeBy } from 'fkit'
 * removeBy((a, b) => a === b, 2, [1, 2, 3]) // [1, 3]
 * removeBy((a, b) => a === b, 'f', 'foo') // 'oo'
 */

var removeBy$1 = curry(removeBy);

function fold(f, s, as) {
  return Array.from(as).reduce(f, s);
}

function differenceBy(f, as, bs) {
  return fold(flip(removeBy$1(f)), nub(as), nub(bs));
}

/**
 * Calculates the difference of two lists.
 *
 * The comparator function `f` compares two elements, `a` and `b`. If the
 * elements are considered to be equal, then the comparator function should
 * return `true`. Otherwise it should return `false`.
 *
 * @function
 * @param {Function} f The comparator function.
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array|String} A list that contains the difference of the elements
 * in the lists of `as` and `bs`.
 * @example
 *
 * import { differenceBy } from 'fkit'
 * differenceBy((a, b) => a === b, [1, 2, 3], [2, 3, 4]) // [1]
 * differenceBy((a, b) => a === b, 'hello', 'world') // 'he'
 */

var differenceBy$1 = curry(differenceBy);

/**
 * Calculates the difference of two lists.
 *
 * This is a special case of the `differenceBy` function where the elements are
 * compared using the strict equality `===` operator.
 *
 * @function
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array|String} A list that contains the difference of the elements
 * in the lists of `as` and `bs`.
 * @example
 *
 * import { difference } from 'fkit'
 * difference([1, 2, 3], [2, 3, 4]) // [1]
 * difference('hello', 'world') // 'he'
 */

var difference = differenceBy$1(eq$1);

function div(a, b) {
  return b / a;
}

/**
 * Returns the division of two numbers.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Number} The result of `b / a`.
 * @example
 *
 * import { div } from 'fkit'
 * div(2, 10) // 5
 */

var div$1 = curry(div);

function drop(n, as) {
  var s = isString(as) ? '' : [];
  var m = as.length;

  for (var i = n; i < m; i++) {
    s = s.concat(as[i]);
  }

  return s;
}

/**
 * Drops a number of elements from the start of a list.
 *
 * @function
 * @param {Number} n The number of elements to drop.
 * @param {Array|String} as The list.
 * @returns {Array|String} The result after dropping `n` elements from the list
 * of `as`.
 * @example
 *
 * import { drop } from 'fkit'
 * drop(2, [1, 2, 3]) // [3]
 * drop(2, 'foo') // 'o'
 */

var drop$1 = curry(drop);

/**
 * Drops elements from the start of a list using a predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that drops elements from the list of `as`
 * while the predicate function `p` is satisfied.
 * @example
 *
 * import { dropWhile, lt, neq } from 'fkit'
 * dropWhile(lt(3), [1, 2, 3]) // [3]
 * dropWhile(neq(o), 'foo') // 'oo'
 */

var dropWhile$1 = curry(dropWhile);

function elem(a, as) {
  return as.indexOf(a) >= 0;
}

/**
 * Determines if a value is present in a list.
 *
 * @function
 * @param a The value to find.
 * @param {Array|String} as The list.
 * @returns {Boolean} `true` if the list of `as` contains the value `a`,
 * `false` otherwise.
 * @example
 *
 * import { elem } from 'fkit'
 *
 * elem(0, [1, 2, 3]) // false
 * elem(1, [1, 2, 3]) // true
 *
 * elem('a', 'foo') // false
 * elem('o', 'foo') // true
 */

var elem$1 = curry(elem);

function elemIndex(a, as) {
  var i = as.indexOf(a);
  return i >= 0 ? i : undefined;
}

/**
 * Gets the index of the first occurance of a value in a list.
 *
 * @function
 * @param a The value to find.
 * @param {Array|String} as The list.
 * @returns {Number} The index of the first occurance of the value `a` in the
 * list of `as`, or `undefined` if no value was found.
 * @example
 *
 * import { elemIndex } from 'fkit'
 *
 * elemIndex(0, [1, 2, 3]) // undefined
 * elemIndex(1, [1, 2, 3]) // 0
 *
 * elemIndex('a', 'foo') // undefined
 * elemIndex('o', 'foo') // 1
 */

var elemIndex$1 = curry(elemIndex);

function findIndices(p, as) {
  var s = [];
  var n = as.length;

  for (var i = 0; i < n; i++) {
    if (p(as[i])) {
      s.push(i);
    }
  }

  return s;
}

function elemIndices(a, as) {
  return findIndices(eq$1(a), as);
}

/**
 * Gets the indices of all occurances of a value in a list.
 *
 * @function
 * @param a The value to find.
 * @param {Array|String} as The list.
 * @returns {Array} The indices of all occurances of the value `a` in the list
 * of `as`.
 * @example
 *
 * import { elemIndices } from 'fkit'
 *
 * elemIndices(0, [1, 2, 3]) // []
 * elemIndices(1, [1, 2, 3]) // [0]
 *
 * elemIndices('a', 'foo') // []
 * elemIndices('o', 'foo') // [1, 2]
 */

var elemIndices$1 = curry(elemIndices);

/**
 * Determines whether a number is even.
 *
 * @param {Number} a The number.
 * @returns {Boolean} `true` if the number `a` is even, false otherwise.
 * @example
 *
 * import { even } from 'fkit'
 * even(2) // true
 * even(1) // false
 */
function even(a) {
  return !(a % 2);
}

/**
 * Filters a list using a predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list to filter.
 * @returns {Array|String} A list that contains only the elements in the list
 * of `as` that satisfy the predicate function `p`.
 * @example
 *
 * import { eq, filter, gt } from 'fkit'
 * filter(gt(1), [1, 2, 3]) // [2, 3]
 * filter(eq('o'), 'foo') // 'oo'
 */

var filter$1 = curry(filter);

function find(p, as) {
  return head(filter(p, as));
}

/**
 * Finds the first value in a list that satisfies a predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns A value in the list of `as` that satisfies the predicate function
 * `p`, or `undefined` if no value was found.
 * @example
 *
 * import { eq, find, gt } from 'fkit'
 *
 * find(gt(1), []) // undefined
 * find(gt(1), [1, 2, 3]) // 2
 *
 * find(eq('o'), '') // undefined
 * find(eq('o'), 'foo') // 'o'
 */

var find$1 = curry(find);

function findIndex(p, as) {
  var n = as.length;

  for (var i = 0; i < n; i++) {
    if (p(as[i])) {
      return i;
    }
  }

  return undefined;
}

/**
 * Finds the index of the first occurance of a value in a list that satisfies a
 * predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns {Number} An index of the first occurance of a value in the list of
 * `as` that satisfies the predicate function `p`, or `undefined` if no value
 * was found.
 * @example
 *
 * import { eq, findIndex, gt } from 'fkit'
 *
 * findIndex(gt(1), []) // undefined
 * findIndex(gt(1), [1, 2, 3]) // 1
 *
 * findIndex(eq('o'), '') // undefined
 * findIndex(eq('o'), 'foo') // 1
 */

var findIndex$1 = curry(findIndex);

/**
 * Finds the indices of all the values in a list that satisfy a predicate
 * function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns {Array} A list of indices that satisfy the predicate function `p`.
 * @example
 *
 * import { eq, findIndices, gt } from 'fkit'
 *
 * findIndices(gt(1), []) // []
 * findIndices(gt(1), [1, 2, 3]) // [1, 2]
 *
 * findIndices(eq('o'), '') // []
 * findIndices(eq('o'), 'foo') // [1, 2]
 */

var findIndices$1 = curry(findIndices);

/**
 * Folds a list from left-to-right with a function.
 *
 * @function
 * @param {Function} f The folding function.
 * @param s The starting value.
 * @param {Array|String} as The list to fold.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` folded left-to-right with the binary function `f` and starting value
 * `s`.
 * @example
 *
 * import { fold } from 'fkit'
 * fold((a, b) => a.concat(b), [], [1, 2, 3]) // [1, 2, 3]
 * fold((a, b) => a.concat(b), '', 'foo') // 'foo'
 */

var fold$1 = curry(fold);

function foldRight(f, s, as) {
  return Array.from(as).reduceRight(flip(f), s);
}

/**
 * Folds a list from right-to-left with a function.
 *
 * @function
 * @param {Function} f The folding function.
 * @param s The starting value.
 * @param {Array|String} as The list to fold.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` folded right-to-left with the binary function `f` and starting value
 * `s`.
 * @example
 *
 * import { foldRight } from 'fkit'
 * foldRight((b, a) => a.concat(b), [], [1, 2, 3]) // [3, 2, 1]
 * foldRight((b, a) => a.concat(b), '', 'foo') // 'oof'
 */

var foldRight$1 = curry(foldRight);

function get(ks, o) {
  ks = isString(ks) ? ks.split('.') : ks;
  return ks.reduce(function (a, k) {
    return a !== undefined ? a[k] : undefined;
  }, o);
}

/**
 * Gets a property of an object using a key path.
 *
 * @function
 * @param {Array|String} ks The key path.
 * @param {Object} o The object.
 * @returns The property at the key path `ks` in the object `o`.
 * @example
 *
 * import { get } from 'fkit'
 * const person = { name: 'Jane', age: 20, address: { city: 'Melbourne', country: 'Australia' } }
 * get('name', person) // 'Jane'
 * get('address.city', person) // 'Melbourne'
 */

var get$1 = curry(get);

function gt(a, b) {
  return b > a;
}

/**
 * Determines whether one number is greater than another.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Boolean} `true` if the number `b` is greater than the number `a`,
 * false otherwise.
 * @example
 *
 * import { gt } from 'fkit'
 * gt(1, 2) // true
 * gt(2, 1) // false
 */

var gt$1 = curry(gt);

function gte(a, b) {
  return b >= a;
}

/**
 * Determines whether one number is greater than or equal to another.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Boolean} `true` if the number `b` is greater than or euqal to the
 * number `a`, false otherwise.
 * @example
 *
 * import { gte } from 'fkit'
 * gte(1, 2) // true
 * gte(2, 1) // false
 * gte(2, 2) // true
 */

var gte$1 = curry(gte);

/**
 * Returns true if `a` is a function.
 *
 * @private
 */
function isFunction(a) {
  return a instanceof Function;
}

function groupBy(f, as) {
  var iteratee = function iteratee(a) {
    return isFunction(f) ? f(a) : get(f, a);
  };

  return fold(function (memo, a) {
    var key = iteratee(a);
    (memo[key] = memo[key] || []).push(a);
    return memo;
  }, {}, as);
}

/**
 * Groups values in a list using a grouping function.
 *
 * The grouping function `f` is applied to each item in the list. It should
 * return a value representing the group that each item belongs to.
 *
 * It is also possible to group a list of objects using a key path, instead of
 * a grouping function. The key path identifies the property to use when
 * grouping the objects in the list.
 *
 * @function
 * @param {Function|String} f The grouping function or key path used to group
 * the values in the list.
 * @param {Array|String} as The list.
 * @returns {Object} An object that contains the groups.
 * @example
 *
 * import { groupBy } from 'fkit'
 * groupBy(Math.floor, [1.1, 2.2, 3.3]) // { '1': [1], '2': [2], '3': [3] }
 * groupBy('length', ['one', 'two', 'three']) // { '3': ['one', 'two'], '5': ['three'] }
 * groupBy('a.b', [{ a: { b: 'foo' } }, { a: { b: 'bar' } }]) // { 'foo': [{ a: { b: 'foo' } }], 'bar': [{ a: { b: 'bar' } }] }
 */

var groupBy$1 = curry(groupBy);

/**
 * Increments a number.
 *
 * @param {Number} a The number.
 * @returns {Number} The result of `a + 1`.
 * @example
 *
 * import { inc } from 'fkit'
 * inc(1) // 2
 */
function inc(a) {
  return a + 1;
}

/**
 * Gets the elements before the last element in a list.
 *
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the elements before the last
 * element in the list of `as`.
 * @example
 *
 * import { init } from 'fkit'
 * init([1, 2, 3]) // [1, 2]
 * init('foo') // 'fo'
 */
function init(as) {
  return as.slice(0, as.length - 1);
}

/**
 * Prepends a value to a list.
 *
 * @function
 * @param a The value to prepend.
 * @param {Array|String} bs The list.
 * @returns {Array|String} A list that contains the value `a` prepended to the
 * list of `bs`.
 * @example
 *
 * import { prepend } from 'fkit'
 * prepend(1, [2, 3]) // [1, 2, 3]
 * prepend('f', 'oo') // 'foo'
 */

var prepend$1 = curry(prepend);

/**
 * Gets all initial segments of a list.
 *
 * @param {Array|String} as The list.
 * @returns {Array} A list that contains all initial segments of the list of
 * `as`.
 * @example
 *
 * import { inits } from 'fkit'
 * inits([1, 2, 3]) // [[], [1], [1, 2], [1, 2, 3]]
 * inits('foo') // ['', 'f', 'fo', 'foo']
 */

function inits(as) {
  return prepend$1(mempty(as), empty(as) ? [] : inits(tail(as)).map(prepend$1(head(as))));
}

function intersectBy(f, as, bs) {
  return fold(function (cs, a) {
    return any(function (b) {
      return f(a, b);
    }, bs) ? append(a, cs) : cs;
  }, mempty(as), as);
}

/**
 * Calculates the intersection of two lists.
 *
 * The comparator function `f` compares two elements, `a` and `b`. If the
 * elements are considered to be equal, then the comparator function should
 * return `true`. Otherwise it should return `false`.
 *
 * Duplicates are removed from `bs`, but if `as` contains duplicates then so
 * will the result.
 *
 * @function
 * @param {Function} f The comparator function.
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array|String} A list that contains the intersection of the elments
 * in the lists of `as` and `bs`.
 * @example
 *
 * import { intersectBy } from 'fkit'
 * intersectBy((a, b) => a === b, [1, 2, 3], [2, 3, 4]) // [2, 3]
 * intersectBy((a, b) => a === b, 'hello', 'world') // 'ol'
 */

var intersectBy$1 = curry(intersectBy);

/**
 * Calculates the intersection of two lists.
 *
 * This is a special case of the `intersectBy` function where the elements are
 * compared using the strict equality `===` operator.
 *
 * Duplicates are removed from `bs`, but if `as` contains duplicates then so
 * will the result.
 *
 * @function
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array|String} A list that contains the intersection of the elments
 * in the lists of `as` and `bs`.
 * @example
 *
 * import { intersect } from 'fkit'
 * intersect([1, 2, 3], [2, 3, 4]) // [2, 3]
 * intersect('hello', 'world') // 'ol'
 */

var intersect = intersectBy$1(eq$1);

function intersperse(s, as) {
  var prependToAll = function prependToAll(bs) {
    return empty(bs) ? mempty(bs) : concat(s, head(bs), prependToAll(tail(bs)));
  };

  return empty(as) ? mempty(as) : concat(head(as), prependToAll(tail(as)));
}

/**
 * Intersperses the elements of a list with separator.
 *
 * @function
 * @param {String} s The separator value.
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` interspersed with the separator `s`.
 * @example
 *
 * import { intersperse } from 'fkit'
 * intersperse(4, [1, 2, 3]) // [1, 4, 2, 4, 3]
 * intersperse('-', 'foo') // 'f-o-o'
 */

var intersperse$1 = curry(intersperse);

function isPrefixOf(as, bs) {
  if (empty(as)) {
    return true;
  } else if (empty(bs)) {
    return false;
  } else {
    return head(as) === head(bs) && isPrefixOf(tail(as), tail(bs));
  }
}

/**
 * Determines if a list is a prefix of another list.
 *
 * @function
 * @param {Array|String} as The prefix.
 * @param {Array|String} bs The list to check.
 * @returns {Boolean} `true` if the list of `as` is a prefix of the list of
 * `bs`, `false` otherwise.
 * @example
 *
 * import { isPrefixOf } from 'fkit'
 *
 * isPrefixOf([], [1, 2, 3]) // true
 * isPrefixOf([1, 2], [1, 2, 3]) // true
 * isPrefixOf([2, 3], [1, 2, 3]) // false
 *
 * isPrefixOf('', 'foo') // true
 * isPrefixOf('fo', 'foo') // true
 * isPrefixOf('oo', 'foo') // false
 */

var isPrefixOf$1 = curry(isPrefixOf);

/**
 * Gets all final segments of a list.
 *
 * @param {Array|String} as The list.
 * @returns {Array} A list that contains all final segments of the list of
 * `as`.
 * @example
 *
 * import { tails } from 'fkit'
 * tails([1, 2, 3]) // [[1, 2, 3], [2, 3], [3], []]
 * tails('foo') // ['foo', 'oo', 'o', '']
 */

function tails(as) {
  return prepend(as, empty(as) ? [] : tails(tail(as)));
}

function isInfixOf(as, bs) {
  return any(isPrefixOf$1(as), tails(bs));
}

/**
 * Determines if a list is contained within another list.
 *
 * @function
 * @param {Array|String} as The list to find.
 * @param {Array|String} bs The list.
 * @returns {Boolean} `true` if the list of `as` is contained within the list
 * of `bs`, `false` otherwise.
 * @example
 *
 * import { isInfixOf } from 'fkit'
 *
 * isInfixOf([], [1, 2, 3]) // true
 * isInfixOf([2, 3], [1, 2, 3]) // true
 * isInfixOf([3, 2], [1, 2, 3]) // false
 *
 * isInfixOf('', 'foo') // true
 * isInfixOf('oo', 'foo') // true
 * isInfixOf('of', 'foo') // false
 */

var isInfixOf$1 = curry(isInfixOf);

/**
 * Reverses the elements in a list.
 *
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` in reverse order.
 * @example
 *
 * import { reverse } from 'fkit'
 * reverse([1, 2, 3]) // [3, 2, 1]
 * reverse('foo') // 'oof'
 */

function reverse (as) {
  return Array.from(as).reduce(flip(prepend$1), mempty(as));
}

function isSuffixOf(as, bs) {
  return isPrefixOf(reverse(as), reverse(bs));
}

/**
 * Determines if a list is a suffix of another list.
 *
 * @function
 * @param {Array|String} as The suffix.
 * @param {Array|String} bs The list to check.
 * @returns {Boolean} `true` if the list of `as` is a suffix of the list of
 * `bs`, `false` otherwise.
 * @example
 *
 * import { isSuffixOf } from 'fkit'
 *
 * isSuffixOf([], [1, 2, 3]) // true
 * isSuffixOf([1, 2], [1, 2, 3]) // false
 * isSuffixOf([2, 3], [1, 2, 3]) // true
 *
 * isSuffixOf('', 'foo') // true
 * isSuffixOf('fo', 'foo') // false
 * isSuffixOf('oo', 'foo') // true
 */

var isSuffixOf$1 = curry(isSuffixOf);

/**
 * Gets the keys of an object.
 *
 * @param {Object} o The object.
 * @returns {Array} A list of keys for the properties of the object `o`.
 * @example
 *
 * import { keys } from 'fkit'
 * const person = { name: 'Jane', age: 20, city: 'Melbourne' }
 * keys(person) // ['name', 'age', 'city']
 */
function keys(o) {
  return Object.keys(o);
}

/**
 * Gets the length of a list.
 *
 * @param {Array|String} as The list.
 * @returns {Number} The number of elements in the list of `as`.
 * @example
 *
 * import { length } from 'fkit'
 * length([1, 2, 3]) // 3
 * length('foo') // 3
 */
function length(as) {
  return as.length;
}

function lt(a, b) {
  return b < a;
}

/**
 * Determines whether one number is less than another.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Boolean} `true` if the number `b` is less than the number `a`,
 * false otherwise.
 * @example
 *
 * import { lt } from 'fkit'
 * lt(1, 2) // false
 * lt(2, 1) // true
 */

var lt$1 = curry(lt);

function lte(a, b) {
  return b <= a;
}

/**
 * Determines whether one number is less than or equal to another.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Boolean} `true` if the number `b` is less than or equal to the
 * number `a`, false otherwise.
 * @example
 *
 * import { lte } from 'fkit'
 * lte(1, 2) // false
 * lte(2, 1) // true
 * lte(2, 2) // true
 */

var lte$1 = curry(lte);

/**
 * Maps a function over a list.
 *
 * @function
 * @param {Function} f The function.
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` mapped with the function `f`.
 * @example
 *
 * import { inc, map, toUpper } from 'fkit'
 * map(inc, [1, 2, 3]) // [2, 3, 4]
 * map(toUpper, 'foo') // ['F', 'O', 'O']
 */

var map$1 = curry(map);

/**
 * Determines the largest of two numbers.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Number} The largest of the numbers `a` and `b`.
 * @example
 *
 * import { max } from 'fkit'
 * max(1, 2) // 2
 */

var max$1 = curry(max);

function maximumBy(f, as) {
  return fold(function (a, b) {
    return f(a, b) > 0 ? a : b;
  }, as[0], as);
}

/**
 * Calculates the maximum value of a list using a comparator function.
 *
 * The comparator function compares two elements, `a` and `b`. If `a` is
 * greater than `b`, then the comparator function should return `1`. If `a` is
 * less than `b`, then the comparator function should return `-1`. If both
 * elements are equal then, the comparator function should return `0`.
 *
 * @function
 * @param {Function} f The comparator function.
 * @param {Array|String} as The list.
 * @returns The maximum value in the list of `as` using the comparator function
 * `f`.
 * @example
 *
 * import { compare, maximumBy } from 'fkit'
 * maximumBy(compare, [1, 2, 3]) // 3
 * maximumBy(compare, 'abc') // 'c'
 */

var maximumBy$1 = curry(maximumBy);

/**
 * Calculates the maximum value of a list.
 *
 * This a special case of the `maximumBy` function where the values are
 * compared using natural ordering.
 *
 * @function
 * @param {Array|String} as The list.
 * @returns Returns the maximum value in the list of `as`.
 * @example
 *
 * import { maximum } from 'fkit'
 * maximum([1, 2, 3]) // 3
 * maximum('abc') // 'c'
 */

var maximum = maximumBy$1(compare$1);

/**
 * Determines the smallest of two numbers.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Number} The smallest of the numbers `a` and `b`.
 * @example
 *
 * import { min } from 'fkit'
 * min(1, 2) // 1
 */

var min$1 = curry(min);

function minimumBy(f, as) {
  return fold(function (a, b) {
    return f(a, b) < 0 ? a : b;
  }, as[0], as);
}

/**
 * Calculates the minimum value of a list using a comparator function.
 *
 * The comparator function compares two elements, `a` and `b`. If `a` is
 * greater than `b`, then the comparator function should return `1`. If `a` is
 * less than `b`, then the comparator function should return `-1`. If both
 * elements are equal, then the comparator function should return `0`.
 *
 * @function
 * @param {Function} f The comparator function.
 * @param {Array|String} as The list.
 * @returns The minimum value in the list of `as` using the comparator function
 * `f`.
 * @example
 *
 * import { compare, minimumBy } from 'fkit'
 * minimumBy(compare, [1, 2, 3]) // 1
 * minimumBy(compare, 'abc') // 'a'
 */

var minimumBy$1 = curry(minimumBy);

/**
 * Calculates the minimum value of a list.
 *
 * This a special case of the `minaximumBy` function where the values are
 * compared using natural ordering.
 *
 * @function
 * @param {Array|String} as The list.
 * @returns Returns the minimum value in the list of `as`.
 * @example
 *
 * import { minimum } from 'fkit'
 * minimum([1, 2, 3]) // 1
 * minimum('abc') // 'a'
 */

var minimum = minimumBy$1(compare$1);

function mod(a, b) {
  return b % a;
}

/**
 * Returns the modulo of two numbers.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Number} The result of `b % a`.
 * @example
 *
 * import { mod } from 'fkit'
 * mod(3, 2) // 1
 */

var mod$1 = curry(mod);

function mul(a, b) {
  return b * a;
}

/**
 * Returns the product of two numbers.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Number} The result of `b * a`.
 * @example
 *
 * import { mul } from 'fkit'
 * mul(2, 3) // 6
 */

var mul$1 = curry(mul);

/**
 * Returns the negation of a number.
 *
 * @param {Number} a The number.
 * @returns {Number} The result.
 * @example
 *
 * import { negate } from 'fkit'
 * negate(1) // -1
 */
function negate(a) {
  return -a;
}

function neq(a, b) {
  return b !== a;
}

/**
 * Determines whether two values are not equal.
 *
 * @function
 * @param a The first value.
 * @param b The second value.
 * @returns {Boolean} `true` if the value `a` is strictly not equal (`!==`) to
 * the value `b`, false otherwise.
 * @example
 *
 * import { neq } from 'fkit'
 * neq(1, 1) // false
 * neq(1, 2) // true
 */

var neq$1 = curry(neq);

/**
 * Returns the logical NOT of a value.
 *
 * @param a The value.
 * @returns {Boolean} `true` if `a` is not truthy, `false` otherwise.
 * @example
 *
 * import { not } from 'fkit'
 * not(true) // false
 * not(false) // true
 */
function not(a) {
  return !a;
}

/**
 * Determines whether a number is odd.
 *
 * @param {Number} a The number.
 * @returns {Boolean} `true` if the number `a` is odd, false otherwise.
 * @example
 *
 * import { odd } from 'fkit'
 * odd(2) // true
 * odd(1) // false
 */
function odd(a) {
  return Boolean(a % 2);
}

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function set_(ks, v, o) {
  var k = head(ks);

  if (ks.length > 1) {
    v = set_(tail(ks), v, o[k] || {});
  }

  return copy(o, _defineProperty({}, k, v));
}

function set(ks, v, o) {
  ks = isString(ks) ? ks.split('.') : ks;
  return set_(ks, v, o);
}

function omit(ks, o) {
  return difference(keys(o), ks).reduce(function (p, k) {
    return set(k, get(k, o), p);
  }, {});
}

/**
 * Omits properties of an object.
 *
 * @function
 * @param {Array} ks The list of keys to omit.
 * @param {Object} o The object.
 * @returns {Object} A copy of the object `o` *without* the properties in the
 * list of `ks`.
 * @example
 *
 * import { omit } from 'fkit'
 * const person = { name: 'Jane', age: 20, city: 'Melbourne' }
 * omit(['name', 'age'], person) // { city: 'Melbourne' }
 */

var omit$1 = curry(omit);

function or(a, b) {
  return Boolean(b || a);
}

/**
 * Returns the logical OR of two values.
 *
 * @function
 * @param a The first value.
 * @param b The second value.
 * @returns {Boolean} The result of `b || a`.
 * @example
 *
 * import { or } from 'fkit'
 * or(false, true) // true
 * or(0, 1) // true
 * or('', 'foo') // true
 */

var or$1 = curry(or);

/**
 * Gets the key-value pairs of an object.
 *
 * @param {Object} o The object.
 * @returns {Array} A list of key-value pairs for the properties of the object
 * `o`.
 * @example
 *
 * import { pairs } from 'fkit'
 * const person = { name: 'Jane', age: 20, city: 'Melbourne' }
 * pairs(person) // [['name', 'Jane'], ['age', 20], ['city', 'Melbourne']]
 */

function pairs(o) {
  return keys(o).map(function (k) {
    return [k, get$1(k, o)];
  });
}

function partition(p, as) {
  return [filter(p, as), filter(compose(not, p), as)];
}

/**
 * Partitions a list using a predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns {Array} A list that contains the elements in the list of `as` split
 * into a pair of lists: the elements that satisfy the predicate function `p`
 * and the elements that do not satisfy the predicate function `p`.
 * @example
 *
 * import { eq, gt, partition } from 'fkit'
 * partition(gt(1), [1, 2, 3]) // [[2, 3], [1]]
 * partition(eq('o'), 'foo') // ['oo', 'f']
 */

var partition$1 = curry(partition);

function permutations_(bs_, cs) {
  var b = head(bs_);
  var bs = tail(bs_);
  return empty(bs_) ? [] : foldRight(interleave, permutations_(bs, prepend$1(b, cs)), permutations(cs));

  function interleave(ds, r) {
    return interleave_(id, ds)[1];

    function interleave_(f, es_) {
      if (empty(es_)) {
        return [bs, r];
      } else {
        var e = head(es_);
        var es = tail(es_);
        var s = interleave_(compose(f, prepend$1(e)), es);
        return [prepend$1(e, s[0]), prepend$1(f(concat(b, e, s[0])), s[1])];
      }
    }
  }
}
/**
 * Calculates the permutations of a list.
 *
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains all the permutations of the
 * elements in the list of `as`.
 * @example
 *
 * import { permutations } from 'fkit'
 * permutations([1, 2, 3]) // [[1, 2, 3], [2, 1, 3], [3, 2, 1], [2, 3, 1], [3, 1, 2], [1, 3, 2]]
 * permutations('abc') // ['abc', 'bac', 'cba', 'bca', 'cab', 'acb']
 */


function permutations(as) {
  return prepend$1(as, permutations_(as, []));
}

function pick(ks, o) {
  return ks.reduce(function (p, k) {
    return set(k, get(k, o), p);
  }, {});
}

/**
 * Picks properties of an object.
 *
 * @function
 * @param {Array} ks The list of keys.
 * @param {Object} o The object.
 * @returns {Object} A copy of the object `o` *with* the properties in the list
 * of `ks`.
 * @example
 *
 * import { pick } from 'fkit'
 * const person = { name: 'Jane', age: 20, city: 'Melbourne' }
 * pick(['name', 'age'], person) // { name: 'Jane', age: 20 }
 */

var pick$1 = curry(pick);

/**
 * Calculates the product of the elements in a list.
 *
 * @param {Array} as The list.
 * @returns {Number} The product of the elements in the list of `as`.
 * @example
 *
 * import { product } from 'fkit'
 * product([1, 2, 3]) // 6
 */

function product(as) {
  return fold(mul$1, 1, as);
}

function randomFloat(a, b) {
  return Math.random() * (b - a) + a;
}

/**
 * Generates a random float.
 *
 * @function
 * @param {Number} a The minimum number.
 * @param {Number} b The maximum number.
 * @returns {Number} A random float between `a` and `b`.
 * @example
 *
 * import { randomFloat } from 'fkit'
 * randomFloat(1, 3) // 2.3
 */

var randomFloat$1 = curry(randomFloat);

function randomInt(a, b) {
  return Math.floor(Math.random() * (b - a + 1)) + a;
}

/**
 * Generates a random integer.
 *
 * @function
 * @param {Number} a The minimum number.
 * @param {Number} b The maximum number.
 * @returns {Number} A random integer between `a` and `b`.
 * @example
 *
 * import { randomInt } from 'fkit'
 * randomInt(1, 3) // 2
 */

var randomInt$1 = curry(randomInt);

function range(a, n) {
  return array(n).map(function (_, i) {
    return a + i;
  });
}

/**
 * Creates a new array of numbers.
 *
 * @function
 * @param {Number} a The starting number.
 * @param {Number} n The number of numbers to create.
 * @returns {Array} An array of numbers of length `n` starting from `a`.
 * @example
 *
 * import { range } from 'fkit'
 * range(1, 3) // [1, 2, 3]
 */

var range$1 = curry(range);

/**
 * Removes the first occurance of an element from a list.
 *
 * This is a special case of the `removeBy` function where the elements are
 * compared using the strict equality `===` operator.
 *
 * @function
 * @param a The value to remove.
 * @param {Array|String} bs The list.
 * @returns {Array|String} A list with the first occurance of the element `a`
 * removed from the list of `bs`.
 * @example
 *
 * import { remove } from 'fkit'
 * remove(2, [1, 2, 3]) // [1, 3]
 * remove('f', 'foo') // 'oo'
 */

var remove = removeBy$1(eq$1);

function replace(a, b, s) {
  return s.replace(a, b);
}

/**
 * Replaces a term in a string.
 *
 * @function
 * @param {String|RegExp} a The term to find.
 * @param {String} b The string to replace with.
 * @param {String} s The string to perform the find and replace on.
 * @returns {String} The result of replacing term `a` with the string `b` in
 * the string `s`.
 * @example
 *
 * import { replace } from 'fkit'
 * replace('r', 'z', 'bar') // baz
 * replace(/^hello/, 'goodbye', 'hello world!') // goodbye world!
 */

var replace$1 = curry(replace);

/**
 * Creates a new string.
 *
 * @param {Number} n The length of the string.
 * @returns {String} A string of length `n`.
 * @example
 *
 * import { string } from 'fkit'
 * string(3) // '   '
 */

function string(n) {
  return array(n + 1).join(' ');
}

function replicate(n, a) {
  var as = isString(a) ? string(n) : array(n);
  return concatMap(function () {
    return [a];
  }, as);
}

/**
 * Replicates a value.
 *
 * @function
 * @param {Number} n The number of copies.
 * @param a The value to replicate.
 * @returns {Array|String} A list of length `n` with `a` the value of every
 * element.
 * @example
 *
 * import { replicate } from 'fkit'
 * replicate(3, 1) // [1, 1, 1]
 * replicate(3, 'a') // 'aaa'
 */

var replicate$1 = curry(replicate);

/**
 * Shuffles a list using the [Fisher-Yates
 * algorithm](http://en.wikipedia.org/wiki/Fisher-Yates_shuffle).
 *
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` randomly shuffled.
 * @example
 *
 * import { shuffle } from 'fkit'
 * shuffle([1, 2, 3]) // [2, 3, 1]
 * shuffle('abc') // 'bca'
 */

function shuffle(as) {
  var i = -1;

  var f = function f(b, a) {
    var j = randomInt$1(0, ++i);
    b[i] = b[j];
    b[j] = a;
    return b;
  };

  var r = array(as.length);
  var bs = fold(f, r, as);
  var s = isString(as) ? '' : [];
  return concatWith(s, bs);
}

function take(n, as) {
  var s = isString(as) ? '' : [];
  var m = as.length;

  for (var i = 0; i < Math.min(m, n); i++) {
    s = s.concat(as[i]);
  }

  return s;
}

function sample(n, as) {
  return take(n, shuffle(as));
}

/**
 * Samples random elements from a list.
 *
 * @function
 * @param {Number} n The number of elements to sample.
 * @param {Array|String} as The list.
 * @returns {Array|String} A list of `n` elements randomly sampled from the
 * list of `as`.
 * @example
 *
 * import { sample } from 'fkit'
 * sample(2, [1, 2, 3]) // [3, 1]
 * sample(2, 'abc') // 'ca'
 */

var sample$1 = curry(sample);

function tap(f, a) {
  f(a);
  return a;
}

function scan(f, s, as) {
  var r = [s];
  fold(function (b, a) {
    return tap(r.push.bind(r), f(b, a));
  }, s, as);
  return r;
}

/**
 * Scans a list from left-to-right with a function.
 *
 * @function
 * @param {Function} f The folding function.
 * @param s The starting value.
 * @param {Array|String} as The list to scan.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` scanned left-to-right with the binary function `f` and starting value
 * `s`.
 * @example
 *
 * import { scan } from 'fkit'
 * scan((a, b) => a.concat(b), [],  [1, 2, 3]) // [[], [1], [1, 2], [1, 2, 3]]
 * scan((a, b) => a.concat(b), '',  'foo') // ['', 'f', 'fo', 'foo']
 */

var scan$1 = curry(scan);

function scanRight(f, s, as) {
  var r = [s];
  foldRight(function (a, b) {
    return tap(r.unshift.bind(r), f(a, b));
  }, s, as);
  return r;
}

/**
 * Scans a list from right-to-left with a function.
 *
 * @function
 * @param {Function} f The folding function.
 * @param s The starting value.
 * @param {Array|String} as The list to scan.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` scanned right-to-left with the binary function `f` and starting value
 * `s`.
 * @example
 *
 * import { scanRight } from 'fkit'
 * scanRight((b, a) => a.concat(b), [],  [1, 2, 3]) // [[3, 2, 1], [3, 2], [3], []]
 * scanRight((b, a) => a.concat(b), '',  'foo') // ['oof', 'oo', 'o', '']
 */

var scanRight$1 = curry(scanRight);

/**
 * Sets a property of an object using a key path.
 *
 * @function
 * @param {Array|String} ks The key path.
 * @param v The value to set.
 * @param {Object} o The object.
 * @returns {Object} A copy of the object `o` with the property at the key
 * path `ks` set to the value `v`.
 * @example
 *
 * import { set } from 'fkit'
 * const person = { name: 'Jane', age: 20, address: { city: 'Melbourne', country: 'Australia' } }
 * set('name', 'Steve', person) // { name: 'Steve', age: 20, address: { city: 'Melbourne', country: 'Australia' } }
 * set('address.city', 'Sydney', person) // { name: 'Jane', age: 20, address: { city: 'Sydney', country: 'Australia' } }
 */

var set$1 = curry(set);

/**
 * Converts the list of `as` to a list of type `t`.
 *
 * @private
 */
function toList(as, t) {
  return t === 'string' ? as.join('') : as;
}

function sortBy(c, as) {
  var bs = Array.from(as.slice(0));
  return toList(bs.sort(c), _typeof(as));
}

/**
 * Sorts a list using a comparator function.
 *
 * The comparator function compares two elements, `a` and `b`. If `a` is
 * greater than `b`, then the comparator function should return `1`. If `a` is
 * less than `b`, then the comparator function should return `-1`. If both
 * elements are equal then, the comparator function should return `0`.
 *
 * @function
 * @param {Function} c The comparator function.
 * @param {Array|String} as The list to sort.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` sorted using the comparator function `c`.
 * @example
 *
 * import { compare, sortBy } from 'fkit'
 * sortBy(compare, [2, 3, 1]) // [1, 2, 3]
 * sortBy(compare, 'bca') // 'abc'
 */

var sortBy$1 = curry(sortBy);

/**
 * Sorts a list using natural ordering.
 *
 * This a special case of the `sortBy` function where the values are compared
 * using natural ordering.
 *
 * @function
 * @param {Array|String} a The list to sort.
 * @returns {Array|String} A list that contains the elements in the list of
 * `as` sorted.
 * @example
 *
 * import { sort } from 'fkit'
 * sort([2, 3, 1]) // [1, 2, 3]
 * sort('bca') // 'abc'
 */

var sort = sortBy$1(compare);

function splitAt(n, as) {
  return [take(n, as), drop(n, as)];
}

/**
 * Splits a list.
 *
 * @function
 * @param {Number} n The length of the prefix.
 * @param {Array|String} as The list.
 * @returns {Array} A list that contains the elements in the list of `as` split
 * into a pair of lists: a prefix of length `n` and the remainder of the list.
 * @example
 *
 * import { splitAt } from 'fkit'
 * splitAt(1, [1, 2, 3]) // [[1], [2, 3]]
 * splitAt(1, 'foo') // ['f', 'oo']
 */

var splitAt$1 = curry(splitAt);

function sub(a, b) {
  return b - a;
}

/**
 * Returns the difference of two numbers.
 *
 * @function
 * @param {Number} a The first number.
 * @param {Number} b The second number.
 * @returns {Number} The result of `b - a`.
 * @example
 *
 * import { sub } from 'fkit'
 * sub(1, 2) // 1
 */

var sub$1 = curry(sub);

/**
 * Returns `a` in a pure context.
 *
 * @private
 */

function pure(a) {
  return isString(a) || isArrayOfStrings(a) ? a : [a];
}

/**
 * Calculates the subsequences of a list.
 *
 * @param {Array|String} as The list.
 * @returns {Array} A list that contains all the subsequences of the elements
 * in the list of `as`.
 * @example
 *
 * import { subsequences } from 'fkit'
 * subsequences([1, 2, 3]) // [[], [1], [2], [1, 2], [3], [1, 3], [2, 3], [1, 2, 3]]
 * subsequences('abc') // ['', 'a', 'b', 'ab', 'c', 'ac', 'bc', 'abc']
 */

function subsequences(as) {
  var subsequences_ = function subsequences_(bs) {
    var b = head(bs);

    var f = function f(ys, r) {
      return concat(pure(ys), pure(prepend$1(b, ys)), r);
    };

    if (empty(bs)) {
      return [];
    } else {
      return prepend$1(pure(b), foldRight(f, [], subsequences_(tail(bs))));
    }
  };

  return prepend$1(mempty(as), subsequences_(as));
}

/**
 * Calculates the sum of the elements in a list.
 *
 * @param {Array} as The list.
 * @returns {Number} The sum of the elements in the list of `as`.
 * @example
 *
 * import { sum } from 'fkit'
 * sum([1, 2, 3]) // 6
 */

function sum(as) {
  return fold(add, 0, as);
}

function surround(a, b, cs) {
  return append(b, prepend(a, cs));
}

/**
 * Surrounds the list of `cs` with the values `a` and `b`.
 *
 * @function
 * @param a The left value.
 * @param b The right value.
 * @param {Array|String} cs The list to surround.
 * @returns {Array|String} A new list.
 * @example
 *
 * import { surround } from 'fkit'
 * surround(0, 4, [1, 2, 3]) // [0, 1, 2, 3, 4]
 * surround('(', ')', 'foo') // '(foo)'
 */

var surround$1 = curry(surround);

/**
 * Takes a number of elements from the start of a list.
 *
 * @function
 * @param {Number} n The number of elements to take.
 * @param {Array|String} as The list.
 * @returns {Array|String} The result after taking `n` elements from the list
 * of `as`.
 * @example
 *
 * import { take } from 'fkit'
 * take(2, [1, 2, 3]) // [1, 2]
 * take(2, 'foo') // 'fo'
 */

var take$1 = curry(take);

/**
 * Gets the prefix of a list using a predicate function.
 *
 * @function
 * @param {Function} p The predicate function.
 * @param {Array|String} as The list.
 * @returns {Array|String} A list that takes the elements from the list of `as`
 * while the predicate function `p` is satisfied.
 * @example
 *
 * import { lt, neq, takeWhile } from 'fkit'
 * takeWhile(lt(3), [1, 2, 3]) // [1, 2]
 * takeWhile(neq(o), 'foo') // 'f'
 */

var takeWhile$1 = curry(takeWhile);

/**
 * Applies the function `f` to the value `a` and returns the value `a`
 * unchanged.
 *
 * @function
 * @param {Function} f The function to apply.
 * @param a The value.
 * @returns The value `a` unchanged.
 * @example
 *
 * import { tap } from 'fkit'
 * const f = a => console.log(a)
 * tap(f)(1) // 1
 */

var tap$1 = curry(tap);

/**
 * Converts a string to lowercase.
 *
 * @param {String} s The string.
 * @returns {String} A new string.
 * @example
 *
 * import { toLower } from 'fkit'
 * toLower('FOO') // foo
 */
function toLower(s) {
  return s.toLowerCase();
}

/**
 * Converts a string to uppercase.
 *
 * @param {String} s The string.
 * @returns {String} A new string.
 * @example
 *
 * import { toUpper } from 'fkit'
 * toUpper('foo') // FOO
 */
function toUpper(s) {
  return s.toUpperCase();
}

/**
 * Creates a new tuple.
 *
 * @param as The values.
 * @returns {Array} A tuple with the values from `as`.
 * @example
 *
 * import { tuple } from 'fkit'
 * tuple(1, 2, 3) // [1, 2, 3]
 * tuple('a', 'b', 'c') // ['a', 'b', 'c']
 */
function tuple() {
  for (var _len = arguments.length, as = new Array(_len), _key = 0; _key < _len; _key++) {
    as[_key] = arguments[_key];
  }

  return as;
}

/**
 * Converts a function of any arity to a unary function.
 *
 * @param {Function} f The function.
 * @returns {Function} A function that wraps the function `f` to accept only
 * one argument.
 * @example
 *
 * import { unary } from 'fkit'
 * function f (a) { ... }
 * const g = unary(f)
 * g(1, 2, 3) // f(1)
 */

function unary(f) {
  return f.length === 1 ? f : apply$1(f);
}

/**
 * Converts a binary function to a function on pairs.
 *
 * @param {Function} f The function.
 * @returns {Function} A function that wraps the binary function `f` to accept
 * a pair.
 * @example
 *
 * import { uncurry } from 'fkit'
 * const add = uncurry((a, b) => a + b)
 * add([1, 2]) // 3
 */
function uncurry(f) {
  return function (p) {
    return f(p[0], p[1]);
  };
}

function unionBy(f, as, bs) {
  return fold(function (cs, b) {
    return any(function (a) {
      return f(a, b);
    }, as) ? cs : append(b, cs);
  }, as, bs);
}

/**
 * Calculates the union of two lists.
 *
 * The comparator function `f` compares two elements, `a` and `b`. If the
 * elements are both considered to equal, then the comparator function should
 * return `true`. Otherwise it should return `false`.
 *
 * Duplicates are removed from `bs`, but if `as` contains duplicates then so
 * will the result.
 *
 * @function
 * @param {Function} f The comparator function.
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array|String} A list that contains the union of elements in the
 * lists of `as` and `bs`.
 * @example
 *
 * import { unionBy } from 'fkit'
 * unionBy((a, b) => a === b, [1, 2, 3], [2, 3, 4]) // [1, 2, 3, 4]
 * unionBy((a, b) => a === b, 'hello', 'world') // 'hellowrd'
 */

var unionBy$1 = curry(unionBy);

/**
 * Calculates the union of two lists.
 *
 * This is a special case of the `unionBy` function where the elements are
 * compared using the strict equality `===` operator.
 *
 * Duplicates are removed from `bs`, but if `as` contains duplicates then so
 * will the result.
 *
 * @function
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array|String} A list that contains the union of elements in the
 * lists of `as` and `bs`.
 * @example
 *
 * import { union } from 'fkit'
 * union([1, 2, 3], [2, 3, 4]) // [1, 2, 3, 4]
 * union('hello', 'world') // 'hellowrd'
 */

var union = unionBy$1(eq$1);

/**
 * Unzips a list of pairs into a pair of lists.
 *
 * @param {Array} as The list.
 * @returns {Array} The list of pairs `as` unzipped into a pair of lists.
 * @example
 *
 * import { unzip } from 'fkit'
 * unzip([[1, 4], [2, 5], [3, 6]]) // [[1, 2, 3], [4, 5, 6]]
 * unzip([['f', 'b'], ['o', 'a'], ['o', 'r']]) // ['foo', 'bar']
 */

function unzip(as) {
  var s = mempty(as[0]);
  return as.reduceRight(function (p, ps) {
    var a = ps[0];
    var b = ps[1];
    var as = p[0];
    var bs = p[1];
    return [prepend(a, as), prepend(b, bs)];
  }, [s, s]);
}

function update(ks, f, o) {
  return set(ks, f(get(ks, o)), o);
}

/**
 * Updates a property of an object with a function.
 *
 * @function
 * @param {Array|String} ks The key path.
 * @param {Function} f The function.
 * @param {Object} o The object.
 * @returns {Object} A copy of the object `o` with the property at the key
 * path `ks` updated with the function `f`.
 * @example
 *
 * import { always, inc, update } from 'fkit'
 * const person = { name: 'Jane', age: 20, address: { city: 'Melbourne', country: 'Australia' } }
 * update('age', inc, person) // { name: 'Jane', age: 21, address: { city: 'Melbourne', country: 'Australia' } }
 * update('address.city', always('Sydney'), person) // { name: 'Jane', age: 20, address: { city: 'Sydney', country: 'Australia' } }
 */

var update$1 = curry(update);

/**
 * Gets the values of an object.
 *
 * @param {Object} o The object.
 * @returns {Array} A list of values for the properties of the object `o`.
 * @example
 *
 * import { values } from 'fkit'
 * const person = { name: 'Jane', age: 20, city: 'Melbourne' }
 * values(person) // ['Jane', 20, 'Melbourne']
 */

function values(o) {
  return keys(o).map(function (k) {
    return flip(get)(o, k);
  });
}

function whereAll(ps, a) {
  return ps.map(applyRight$1(a)).reduce(and$1, true);
}

/**
 * Determines whether all predicate functions in a list are satisfied they are
 * applied to a value.
 *
 * @function
 * @param {Array} ps The list of predicate functions.
 * @param a The value.
 * @returns {Boolean} `true` if all predicate functions in the list of `ps` are
 * satisfied when applied to the value `a`, `false` otherwise.
 * @example
 *
 * import { gt, whereAll } from 'fkit'
 * const ps = [gt(1), gt(2)]
 * whereAll(ps, 1) // false
 * whereAll(ps, 2) // false
 * whereAll(ps, 3) // true
 */

var whereAll$1 = curry(whereAll);

function whereAny(ps, a) {
  return ps.map(applyRight$1(a)).reduce(or$1, false);
}

/**
 * Determines whether any predicate functions in a list are satisfied when they
 * are applied to a value.
 *
 * @function
 * @param {Array} ps The list of predicate functions.
 * @param a The value.
 * @returns {Boolean} `true` if any predicate function in the list of `ps` are
 * satisfied when applied to the value `a`, `false` otherwise.
 * @example
 *
 * import { gt, whereAny } from 'fkit'
 * const ps = [gt(1), gt(2)]
 * whereAny(ps, 1) // false
 * whereAny(ps, 2) // true
 * whereAny(ps, 3) // true
 */

var whereAny$1 = curry(whereAny);

function zipWith(f, as, bs) {
  var n = Math.min(as.length, bs.length);
  return Array.from(as.slice(0, n)).map(function (a, i) {
    return f(a, bs[i]);
  });
}

/**
 * Zips two lists with a function.
 *
 * @function
 * @param {Function} f The zipping function.
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array} The lists of `as` and `bs` zipped with the binary function `f`.
 * @example
 *
 * import { zipWith } from 'fkit'
 * zipWith((a, b) => a + b, [1, 2, 3], [4, 5, 6]) // [5, 7, 9]
 * zipWith((a, b) => a + b, 'foo', 'bar') // ['fb', 'oa', or']
 */

var zipWith$1 = curry(zipWith);

/**
 * Zips two lists into a list of pairs.
 *
 * This is a special case of the `zipWith` function where the elements are
 * combined using the `pair` function.
 *
 * @function
 * @param {Array|String} as The first list.
 * @param {Array|String} bs The second list.
 * @returns {Array} The lists of `as` and `bs` zipped into a list of pairs.
 * @example
 *
 * import { zip } from 'fkit'
 * zip([1, 2, 3], [4, 5, 6]) // [[1, 4], [2, 5], [3, 6]]
 * zip('foo', 'bar') // [['f', 'b'], ['o', 'a'], ['o', 'r']]
 */

var zip = zipWith$1(pair$1);

export { add$1 as add, all$1 as all, always, and$1 as and, any$1 as any, append$1 as append, apply$1 as apply, apply2$1 as apply2, apply3$1 as apply3, applyMethod$1 as applyMethod, applyMethod2$1 as applyMethod2, applyMethod3$1 as applyMethod3, applyRight$1 as applyRight, array, between$1 as between, binary, branch$1 as branch, cartesian$1 as cartesian, chunk, chunkBy$1 as chunkBy, clamp$1 as clamp, compare$1 as compare, compose, concat, concatMap$1 as concatMap, copy, curry, dec, difference, differenceBy$1 as differenceBy, div$1 as div, drop$1 as drop, dropWhile$1 as dropWhile, elem$1 as elem, elemIndex$1 as elemIndex, elemIndices$1 as elemIndices, empty, eq$1 as eq, even, filter$1 as filter, find$1 as find, findIndex$1 as findIndex, findIndices$1 as findIndices, flip, fold$1 as fold, foldRight$1 as foldRight, get$1 as get, groupBy$1 as groupBy, gt$1 as gt, gte$1 as gte, head, id, inc, init, inits, intersect, intersectBy$1 as intersectBy, intersperse$1 as intersperse, isInfixOf$1 as isInfixOf, isPrefixOf$1 as isPrefixOf, isSuffixOf$1 as isSuffixOf, keys, last, length, lt$1 as lt, lte$1 as lte, map$1 as map, max$1 as max, maximum, maximumBy$1 as maximumBy, min$1 as min, minimum, minimumBy$1 as minimumBy, mod$1 as mod, mul$1 as mul, negate, neq$1 as neq, not, nub, nubBy$1 as nubBy, odd, omit$1 as omit, or$1 as or, pair$1 as pair, pairs, partition$1 as partition, permutations, pick$1 as pick, prepend$1 as prepend, product, randomFloat$1 as randomFloat, randomInt$1 as randomInt, range$1 as range, remove, removeBy$1 as removeBy, replace$1 as replace, replicate$1 as replicate, reverse, sample$1 as sample, scan$1 as scan, scanRight$1 as scanRight, set$1 as set, shuffle, sort, sortBy$1 as sortBy, span$1 as span, splitAt$1 as splitAt, string, sub$1 as sub, subsequences, sum, surround$1 as surround, tail, tails, take$1 as take, takeWhile$1 as takeWhile, tap$1 as tap, toLower, toUpper, tuple, unary, uncurry, union, unionBy$1 as unionBy, unzip, update$1 as update, values, variadic, whereAll$1 as whereAll, whereAny$1 as whereAny, zip, zipWith$1 as zipWith };
